from typing import Optional
import click


def echo_error(msg: str):
    click.echo("\x1b(B\x1b[m[\x1b[33m✕\x1b(B\x1b[m] {error}: {message}".format(
        error=click.style('Error', fg='red'),
        message=msg))


def echo_success(msg: Optional[str] = None, only_msg=False):
    if msg is None:
        msg = "Done"

    if not only_msg:
        msg = "\x1b(B\x1b[m[\x1b[33m✓\x1b(B\x1b[m] " + str(msg)

    click.echo(msg)



from inquirer.themes import Default
from blessed import Terminal


term = Terminal()


class ModifiedTheme(Default):
    def __init__(self):
        super(ModifiedTheme, self).__init__()
        self.List.selection_color = term.bold
        self.List.selection_cursor = ">"
        self.List.unselected_color = term.normal



import click

from ._base_cmd import cli, SharedObject
from .._utils import CustomClickException, with_whitelist, click_ton_exception_handler, click_echo_success
from ..._utils import md_table
from wton.tonclient.utils import WhitelistContactAlreadyExistsError, WhitelistContactDoesNotExistError
from wton.tonclient import ton_exceptions_handler
from wton.tonsdk.utils import InvalidAddressError, TonCurrencyEnum, Address


@cli.group()
def whitelist():
    """
    Operate with whitelist contacts.
    """


@whitelist.command()
@with_whitelist
@click.argument('name', required=True)
@click.argument('address', required=True)
@click.pass_obj
def add(shared_object: SharedObject, name, address):
    """
    Add a contact to the whitelist.
    """
    try:
        shared_object.whitelist.add_contact(name, address, save=True)
    except (PermissionError, FileNotFoundError, WhitelistContactAlreadyExistsError, InvalidAddressError) as e:
        raise CustomClickException(e)

    click_echo_success(f"contact {name} has been added to the whitelist.")

@whitelist.command()
@with_whitelist
@click.argument('name', required=True)
@click.pass_obj
def get(shared_object: SharedObject, name):
    """
    Get a contact address by its name.
    """
    try:
        contact = shared_object.whitelist.get_contact(name, raise_none=True)
    except WhitelistContactDoesNotExistError as e:
        raise CustomClickException(e)

    addr = Address(contact.address)
    click.echo(f"Raw address: {addr.to_string(False, False, False)}")
    click.echo(f"Nonbounceable address: {addr.to_string(True, True, False)}")
    click.echo(f"Bounceable address: {addr.to_string(True, True, True)}")


@whitelist.command()
@with_whitelist
@click.argument('name', required=True, metavar="EXISTING_CONTACT_NAME")
@click.option('--name', '-n', 'new_name', help="Set a new name.")
@click.option('--address', '-a', 'new_address', help="Set a new address.")
@click.pass_obj
def edit(shared_object: SharedObject, name: str, new_name: str, new_address: str):
    """
    Edit contact in a whitelist.
    """
    try:
        shared_object.whitelist.edit_contact(
            name, new_name, new_address, save=True)
    except (WhitelistContactDoesNotExistError, ValueError) as e:
        raise CustomClickException(e)

    click_echo_success(f"contact {name} has been edited.")

@whitelist.command()
@with_whitelist
@click.argument('name', required=True)
@click.pass_obj
def delete(shared_object: SharedObject, name: str):
    """
    Delete contact from a whitelist.
    """
    try:
        shared_object.whitelist.delete_contact(name, save=True)
    except WhitelistContactDoesNotExistError as e:
        raise CustomClickException(e)

    click_echo_success(f"contact {name} has been deleted.")

@whitelist.command()
@ton_exceptions_handler(click_ton_exception_handler)
@with_whitelist
@click.option('--currency', '-c', default=TonCurrencyEnum.ton, show_default=True, type=click.Choice(TonCurrencyEnum))
@click.option('--verbose', '-v', 'verbose', is_flag=True, default=False, help="Extra information from TON.")
@click.pass_obj
def list(shared_object: SharedObject, verbose: bool, currency: TonCurrencyEnum):
    """
    Print all contacts as a markdown table.
    You can output this command into .md file.
    """
    table = md_table()
    field_names = ["Name", "Address"]
    if verbose:
        field_names += ['State', 'Balance']
    table.field_names = field_names

    contacts = shared_object.whitelist.contacts
    if verbose:
        contact_infos = shared_object.ton_client.get_addresses_information(
            [contact.address for contact in contacts], currency)

        for contact, contact_info in zip(contacts, contact_infos):
            table.add_row([contact.name, contact.address,
                          contact_info.state, contact_info.balance])

    else:
        for contact in shared_object.whitelist.contacts:
            table.add_row([contact.name, contact.address])

    click.echo(table)




from enum import Enum
from typing import Optional
from pydantic import BaseModel, validator

from wton import settings


class TonNetworkEnum(str, Enum):
    mainnet = 'mainnet'
    testnet = 'testnet'


class DAppConfig(BaseModel):
    api_key: Optional[str] = None
    network: TonNetworkEnum = TonNetworkEnum.mainnet

    class Config:
        use_enum_values = True
        validate_assignment = True

    @property
    def graphql_url(self):
        if self.network == TonNetworkEnum.mainnet:
            return settings.DAPP_MAINNET_GRAPHQL_URL
        else:
            return settings.DAPP_TESTNET_GRAPHQL_URL

    @property
    def broadcast_url(self):
        if self.network == TonNetworkEnum.mainnet:
            return settings.DAPP_MAINNET_BROADCAST_URL
        else:
            return settings.DAPP_TESTNET_BROADCAST_URL



class WhitelistContactAlreadyExistsError(Exception):
    pass


class WhitelistContactDoesNotExistError(Exception):
    pass


class KeyStoreAlreadyExistsError(Exception):
    pass


class KeyStoreDoesNotExistError(Exception):
    pass


class KeyStoreInvalidPasswordError(Exception):
    pass


class InvalidKeyStoreError(Exception):
    pass


class KeyStoreIsNotSpecifiedError(Exception):
    pass


class RecordAlreadyExistsError(Exception):
    pass


class RecordDoesNotExistError(Exception):
    pass


class InvalidMnemonicsError(Exception):
    pass



import json
import os
import yaml
import platform
from typing import Any, ByteString, Optional
import glob

from .exceptions import StorageError


def global_workdir() -> str:
    home_folder = os.path.expanduser("~")
    platform_name = platform.system().lower()

    if platform_name == "linux" or platform_name == "darwin":
        return os.path.join(home_folder, ".config", "wton")
    else:
        # TODO: implement "windows" platform support
        raise OSError("Your operating system is not supported yet")


def local_workdir(local_dir_name) -> str:
    return os.path.join(os.getcwd(), local_dir_name)


def find_local_workdir(local_dir_name) -> Optional[str]:
    system_root = os.path.abspath(os.sep)
    current_dir = os.getcwd()

    path = os.path.join(current_dir, local_dir_name)
    while current_dir != system_root:
        if os.path.exists(path):
            return path
        current_dir = os.path.abspath(os.path.join(current_dir, os.pardir))
        path = os.path.join(current_dir, local_dir_name)

    return None


def read_yaml(filepath: str):
    with open(filepath) as f:
        try:
            return yaml.safe_load(f)
        except (yaml.scanner.ScannerError, PermissionError, FileNotFoundError) as e:
            raise StorageError(e)


def read_json(filepath: str):
    with open(filepath, "r") as f:
        return json.loads(f.read())


def read_bytes(filepath: str):
    with open(filepath, "rb") as f:
        return f.read()


def save_json(filepath: str, data: Any):
    ensure_parent_dir_exists(filepath)

    if not data:
        data = None

    with open(filepath, 'w') as f:
        json.dump(data, f, ensure_ascii=False, indent=4)


def save_yaml(filepath: str, data: Any):
    ensure_parent_dir_exists(filepath)

    if not data:
        data = None

    try:
        with open(filepath, 'w') as f:
            yaml.safe_dump(data, f, default_flow_style=False)
    except (PermissionError, FileNotFoundError) as e:
        raise StorageError(e)


def save_bytes(filepath: str, data: ByteString):
    ensure_parent_dir_exists(filepath)

    try:
        with open(filepath, 'wb') as f:
            f.write(data)
    except (PermissionError, FileNotFoundError) as e:
        raise StorageError(e)


def ensure_parent_dir_exists(filepath: str):
    ensure_dir_exists(os.path.dirname(filepath))


def ensure_dir_exists(dirpath: str):
    os.makedirs(dirpath, exist_ok=True)


def exists(path: str):
    return os.path.exists(path)


def get_filenames_by_ptrn(dir: str, pattern: str):
    return glob.glob(os.path.join(dir, pattern))



from enum import Enum
from pydantic import BaseModel
from typing import Any, List, Tuple, Union, Generator

from ._wton import WtonConfig
from ._provider import ProviderConfig, DAppConfig, TonNetworkEnum


class ConfigLocation(str, Enum):
    local_location = 'local'
    global_location = 'global'


class Config(BaseModel):
    wton: WtonConfig = WtonConfig()
    provider: ProviderConfig = ProviderConfig()

    class Config:
        use_enum_values = True
        validate_assignment = True

    @classmethod
    def field_names(cls) -> List[str]:
        fields = []
        fields += [f"wton.{field}" for field in list(
            WtonConfig.__fields__)]
        fields += [f"provider.dapp.{field}" for field in list(
            DAppConfig.__fields__)]

        return fields

    def key_value(self, exclude_unset=False) -> Generator[Tuple[str, Any], None, None]:
        for key, val in self.wton.dict(exclude_unset=exclude_unset).items():
            yield f"wton.{key}", val

        for key, val in self.provider.dapp.dict(exclude_unset=exclude_unset).items():
            yield f"provider.dapp.{key}", val

    def get_nondefault_value(self, field: str) -> Union[None, Any]:
        attrs_tree = field.split(".")
        val = self.dict(exclude_unset=True)

        for attr in attrs_tree:
            if attr in val:
                val = val[attr]
            else:
                return None
        return val

    def update_value(self, field: str, value: Any):
        tree = field.split(".")
        field_name = tree[-1]
        attrs = tree[:-1]
        attrs_tree = [self]

        for attr_name in attrs:
            attr = getattr(attrs_tree[-1], attr_name, None)
            attrs_tree.append(attr)

        for idx, attr in enumerate(attrs_tree[::-1], 1):
            setattr(attr, field_name, value)
            value = attr
            if idx <= len(attrs):
                field_name = attrs[-idx]

    def to_nondefault_dict_without_field(self, field: str):
        fields_tree = field.split('.')
        to_exclude = self.__nested_dict({}, fields_tree[:-1], fields_tree[-1])
        return self.dict(exclude_unset=True, exclude=to_exclude)

    def __nested_dict(self, tree, vector, value):
        key = vector[0]
        tree[key] = {value} \
            if len(vector) == 1 \
            else self.__nested_dict(tree[key] if key in tree else {},
                                    vector[1:],
                                    value)
        return tree



from pydantic import BaseModel
from ._dapp import DAppConfig, TonNetworkEnum


class ProviderConfig(BaseModel):
    dapp: DAppConfig = DAppConfig()

    class Config:
        validate_assignment = True



from enum import Enum
from typing import List, Optional, Tuple
from ._wallet_contract_v2 import WalletV2ContractR1, WalletV2ContractR2
from ._wallet_contract_v3 import WalletV3ContractR1, WalletV3ContractR2
from ._wallet_contract_v4 import WalletV4ContractR1, WalletV4ContractR2
from ._wallet_contract import SendModeEnum
from ._wallet_contract import WalletContract
from ...crypto import mnemonic_new, mnemonic_to_wallet_key, mnemonic_is_valid
from ...crypto.exceptions import InvalidMnemonicsError


class WalletVersionEnum(str, Enum):
    v2r1 = 'v2r1'
    v2r2 = 'v2r2'
    v3r1 = 'v3r1'
    v3r2 = 'v3r2'
    v4r1 = 'v4r1'
    v4r2 = 'v4r2'


class Wallets:
    default_version = WalletVersionEnum.v3r2
    ALL = {
        WalletVersionEnum.v2r1: WalletV2ContractR1,
        WalletVersionEnum.v2r2: WalletV2ContractR2,
        WalletVersionEnum.v3r1: WalletV3ContractR1,
        WalletVersionEnum.v3r2: WalletV3ContractR2,
        WalletVersionEnum.v4r1: WalletV4ContractR1,
        WalletVersionEnum.v4r2: WalletV4ContractR2,
    }

    @classmethod
    def create(cls, version: WalletVersionEnum, workchain: int,
               password: Optional[str] = None) -> Tuple[List[str], bytes, bytes, WalletContract]:
        """
        :rtype: (List[str](mnemonics), bytes(public_key), bytes(private_key), WalletContract(wallet))
        """
        mnemonics = mnemonic_new(password=password)
        pub_k, priv_k = mnemonic_to_wallet_key(mnemonics)
        wallet = cls.ALL[version](
            public_key=pub_k, private_key=priv_k, wc=workchain)

        return mnemonics, pub_k, priv_k, wallet

    @classmethod
    def from_mnemonics(cls, mnemonics: List[str], version: WalletVersionEnum = default_version,
                       workchain: int = 0) -> Tuple[List[str], bytes, bytes, WalletContract]:
        """
        :rtype: (List[str](mnemonics), bytes(public_key), bytes(private_key), WalletContract(wallet))
        """
        if not mnemonic_is_valid(mnemonics):
            raise InvalidMnemonicsError()

        pub_k, priv_k = mnemonic_to_wallet_key(mnemonics)
        wallet = cls.ALL[version](
            public_key=pub_k, private_key=priv_k, wc=workchain)

        return mnemonics, pub_k, priv_k, wallet

    @classmethod
    def to_addr_pk(cls, mnemonics: WalletContract, version: WalletVersionEnum = default_version,
                   workchain: int = 0) -> Tuple[bytes, bytes]:
        """
        :rtype: (bytes(addr), bytes(pk)
        """
        _mnemonics, _pub_k, priv_k, wallet = cls.from_mnemonics(
            mnemonics, version, workchain)

        return wallet.address.to_buffer(), priv_k[:32]


__all__ = [
    'WalletV2ContractR1',
    'WalletV2ContractR2',
    'WalletV3ContractR1',
    'WalletV3ContractR2',
    'WalletV4ContractR1',
    'WalletV4ContractR2',
    'WalletContract',
    'SendModeEnum',
    'WalletVersionEnum',
    'Wallets',
]



from ._commands._base_cmd import cli
from ._commands import _config_cmd
from ._commands import _keystore_cmd
from ._commands import _whitelist_cmd
from ._commands import _wallet_cmd
from ._commands import _contract_cmd


def main():
    cli()


from enum import Enum
import os
from typing import Optional
from pydantic import BaseModel, validator

from wton import settings
from wton.tonsdk.contract.wallet import WalletVersionEnum


class TonProviderEnum(str, Enum):
    dapp = "dapp"


class WtonConfig(BaseModel):
    workdir: Optional[str]
    whitelist_path: Optional[str]
    keystore_name: Optional[str] = None
    provider: TonProviderEnum = TonProviderEnum.dapp
    default_wallet_version: WalletVersionEnum = WalletVersionEnum.v3r2

    class Config:
        use_enum_values = True
        validate_assignment = True

    @validator("workdir", always=True)
    def __validate_workdir(cls, v, values):
        if v:
            return v

        return settings.CURRENT_WORKDIR

    @validator("keystore_name", always=True)
    def __validate_keystore_name(cls, v, values):
        if v:
            if not v.endswith(".keystore"):
                v += ".keystore"

        return v

    @property
    def whitelist_path(self):
        return os.path.join(self.workdir, "whitelist.json")

    @property
    def keystores_path(self):
        return os.path.join(self.workdir, "keystores")



import click

from ._base_cmd import cli, SharedObject
from .._utils import CustomClickException, with_whitelist, with_keystore, click_ton_exception_handler
from wton.tonclient import ton_exceptions_handler
from wton.tonclient.utils import RecordDoesNotExistError, WhitelistContactDoesNotExistError


@cli.group()
@with_whitelist
@with_keystore(password_required=False)
@click.option('--address', '-a', help='Show info by an address')
@click.option('--wallet', '-w', help='Show info by a wallet name')
@click.option('--contact', '-c', help='Show info by a whitelist contact name')
@click.pass_obj
def contract(shared_object: SharedObject, address: str, wallet: str, contact: str):
    """
    Operate with contracts.
    """
    if not any([address, wallet, contact]):
        raise click.MissingParameter(
            param=click.Option(['--address', '--wallet', '--contact']))

    if address:
        shared_object.addr = address
    elif wallet:
        try:
            wallet_obj = shared_object.keystore.get_record_by_name(
                wallet, raise_none=True)
        except RecordDoesNotExistError as e:
            raise CustomClickException(e)

        shared_object.addr = wallet_obj.address
    else:
        try:
            contact_obj = shared_object.whitelist.get_contact(
                contact, raise_none=True)
        except WhitelistContactDoesNotExistError as e:
            raise CustomClickException(e)

        shared_object.addr = contact_obj.address


@contract.command()
@ton_exceptions_handler(click_ton_exception_handler)
@click.pass_obj
def info(shared_object: SharedObject):
    """
    Show TON blockchain information.
    """
    addr_info = shared_object.ton_client.get_address_information(
        shared_object.addr)
    for k, v in addr_info.dict().items():
        click.echo(str(k) + ': ' + str(v))


@contract.command()
@ton_exceptions_handler(click_ton_exception_handler)
@click.pass_obj
def balance(shared_object: SharedObject):
    """
    Show the balance of a contract.
    """
    addr_info = shared_object.ton_client.get_address_information(
        shared_object.addr)
    click.echo(addr_info.balance)


@contract.command()
@ton_exceptions_handler(click_ton_exception_handler)
@click.pass_obj
def seqno(shared_object: SharedObject):
    """
    Show seqno.
    """
    seqno = shared_object.ton_client.seqno(shared_object.addr)
    click.echo(seqno)



from ._address import prepare_address, address_state
from ._utils import parse_response
from ._exceptions import ResponseError

all = [
    'prepare_address',
    'address_state',
    'parse_response',
    'ResponseError',
]



from ._whitelist import Whitelist
from ._keystores import KeyStores, KeyStore
from ._exceptions import WhitelistContactAlreadyExistsError, RecordAlreadyExistsError, \
    KeyStoreAlreadyExistsError, InvalidKeyStoreError, WhitelistContactDoesNotExistError, \
    RecordDoesNotExistError, KeyStoreDoesNotExistError, KeyStoreIsNotSpecifiedError, \
    KeyStoreInvalidPasswordError, InvalidMnemonicsError

__all__ = [
    'Whitelist',
    'KeyStores',
    'KeyStore',

    'WhitelistContactAlreadyExistsError',
    'WhitelistContactDoesNotExistError',
    'KeyStoreAlreadyExistsError',
    'KeyStoreDoesNotExistError',
    'KeyStoreIsNotSpecifiedError',
    'KeyStoreInvalidPasswordError',
    'InvalidKeyStoreError',
    'RecordAlreadyExistsError',
    'RecordDoesNotExistError',
    'InvalidMnemonicsError',
]



from typing import Optional
from pydantic import BaseModel, validator

from ._exceptions import WhitelistContactAlreadyExistsError, WhitelistContactDoesNotExistError
from wton.tonsdk.utils import Address, InvalidAddressError
from wton.utils import storage


class WhitelistContact(BaseModel):
    name: str
    address: str

    class Config:
        validate_assignment = True

    @validator('address')
    def validate_address(cls, v, values, **kwargs):
        if v:
            try:
                Address(v)
                return v

            except InvalidAddressError as e:
                raise ValueError(e)


class Whitelist:
    def __init__(self, whitelist_path: str):
        self.whitelist_path = whitelist_path

        try:
            contacts_json = storage.read_json(whitelist_path)
        except FileNotFoundError:
            self.contacts = []
        else:
            if contacts_json:
                self.contacts = [WhitelistContact.parse_obj(contact)
                                 for contact in contacts_json]
            else:
                self.contacts = []

    def add_contact(self, name: str, address: str, save: bool = False):
        address = Address(address).to_string(True, True, True)

        names = {contact.name for contact in self.contacts}
        addresses = {contact.address for contact in self.contacts}

        if name in names:
            raise WhitelistContactAlreadyExistsError(
                f"Contact with the name '{name}' already exists")
        if address in addresses:
            raise WhitelistContactAlreadyExistsError(
                f"Contact with the address {address} already exists")

        self.contacts.append(WhitelistContact(name=name,
                                              address=address))

        if save:
            self.save()

    def get_contact(self, name: str, raise_none: bool = False) -> Optional[WhitelistContact]:
        contact = next(
            (contact for contact in self.contacts if contact.name == name), None)
        if contact is None and raise_none:
            raise WhitelistContactDoesNotExistError(
                f"Contact with the name {name} does not exist")

        return contact

    def edit_contact(self, name: str, new_name: Optional[str], new_address: Optional[str], save: bool = False):
        contact = self.get_contact(name, raise_none=True)
        contact_idx = self.contacts.index(contact)

        if new_name:
            self.contacts[contact_idx].name = new_name

        if new_address:
            self.contacts[contact_idx].address = new_address

        if save:
            self.save()

    def delete_contact(self, name: str, save: bool = False):
        contact = self.get_contact(name, raise_none=True)
        self.contacts.remove(contact)

        if save:
            self.save()

    def save(self):
        contacts = [contact.dict() for contact in self.contacts]
        storage.save_json(self.whitelist_path, contacts)



import decimal
from enum import Enum

from ...boc import Cell
from ...utils import Address, sign_message
from .. import Contract


class SendModeEnum(int, Enum):
    carry_all_remaining_balance = 128
    carry_all_remaining_incoming_value = 64
    destroy_account_if_zero = 32
    pay_gas_separately = 2
    ignore_errors = 1

    def __str__(self) -> str:
        return super().__str__()


class WalletContract(Contract):
    def __init__(self, **kwargs):
        if ("public_key" not in kwargs or "private_key" not in kwargs) and "address" not in kwargs:
            raise Exception(
                "WalletContract required publicKey or address in options")
        super().__init__(**kwargs)

    def create_data_cell(self):
        cell = Cell()
        cell.bits.write_uint(0, 32)
        cell.bits.write_bytes(self.options["public_key"])
        return cell

    def create_signing_message(self, seqno=None):
        seqno = seqno or 0
        cell = Cell()
        cell.bits.write_uint(seqno, 32)
        return cell

    def create_transfer_message(self, to_addr, amount, seqno, payload="",
                                send_mode=SendModeEnum.ignore_errors | SendModeEnum.pay_gas_separately,
                                dummy_signature=False, state_init=None):
        payload_cell = Cell()
        if payload:
            if type(payload) == str:
                payload_cell.bits.write_uint(0, 32)
                payload_cell.bits.write_string(payload)
            elif hasattr(payload, 'refs'):
                payload_cell = payload
            else:
                payload_cell.bits.write_bytes(payload)

        order_header = Contract.create_internal_message_header(
            Address(to_addr), decimal.Decimal(amount))
        order = Contract.create_common_msg_info(
            order_header, state_init, payload_cell)
        signing_message = self.create_signing_message(seqno)
        signing_message.bits.write_uint8(send_mode)
        signing_message.refs.append(order)

        return self.create_external_message(signing_message, seqno, dummy_signature)

    def create_external_message(self, signing_message, seqno, dummy_signature=False):
        signature = bytes(64) if dummy_signature else sign_message(
            bytes(signing_message.bytes_hash()), self.options['private_key']).signature

        body = Cell()
        body.bits.write_bytes(signature)
        body.write_cell(signing_message)

        state_init = code = data = None

        if seqno == 0:
            deploy = self.create_state_init()
            state_init = deploy["state_init"]
            code = deploy["code"]
            data = deploy["data"]

        self_address = self.address
        header = Contract.create_external_message_header(self_address)
        result_message = Contract.create_common_msg_info(
            header, state_init, body)

        return {
            "address": self_address,
            "message": result_message,
            "body": body,
            "signature": signature,
            "signing_message": signing_message,
            "state_init": state_init,
            "code": code,
            "data": data,
        }

    def create_init_external_message(self):
        create_state_init = self.create_state_init()
        state_init = create_state_init["state_init"]
        address = create_state_init["address"]
        code = create_state_init["code"]
        data = create_state_init["data"]

        signing_message = self.create_signing_message()
        signature = sign_message(
            bytes(signing_message.bytes_hash()), self.options['private_key']).signature

        body = Cell()
        body.bits.write_bytes(signature)
        body.write_cell(signing_message)

        header = Contract.create_external_message_header(address)
        external_message = Contract.create_common_msg_info(
            header, state_init, body)

        return {
            "address": address,
            "message": external_message,

            "body": body,
            "signing_message": signing_message,
            "state_init": state_init,
            "code": code,
            "data": data,
        }
